<?php

namespace Drupal\acquia_contenthub\EventSubscriber\SerializeContentField;

use Drupal\acquia_contenthub\AcquiaContentHubEvents;
use Drupal\acquia_contenthub\Event\SerializeCDFEntityField;
use Drupal\Core\Entity\ContentEntityInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Link Field Serializer.
 *
 * This class handles the serialization of menu_link entities.
 */
class LinkFieldSerializer implements EventSubscriberInterface {
  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[AcquiaContentHubEvents::SERIALIZE_CONTENT_ENTITY_FIELD][] = ['onSerializeContentField', 15];
    return $events;
  }

  /**
   * On serialize content field event function.
   * 
   * Extracts entity uuids from link fields and serializes them.
   *
   * @param \Drupal\acquia_contenthub\Event\SerializeCDFEntityField $event
   *   The content entity field serialization event.
   */
  public function onSerializeContentField(SerializeCDFEntityField $event) {
    // Return if the type of field is not a link.
    if ($event->getField()->getFieldDefinition()->getType() != 'link') {
      return;
    }

    // Get main entity.
    $entity = $event->getEntity();

    // Confirm the entity is an instance of ContentEntityInterface.
    if (!$entity instanceof ContentEntityInterface) {
      return;
    }
    $fields = $this->getFieldsFromEntity($entity);
    if (!$fields) {
      return;
    }
    // Get CDF.
    $cdf = $event->getCDF();
    $metadata = $cdf->getMetadata();
    // Init data arr.
    $data = [];

    // Loop through entity fields.
    foreach ($fields as $field) {
      $langcode = $field->getLangcode();

      // Set type in meta data.
      $metadata['field'][$event->getFieldName()] = [
        'type' => $event->getField()->getFieldDefinition()->getType()
      ];

      // Loop through fields to get link values to serialize.
      foreach ($field as $item) {
        // Get values.
        $values = $item->getValue();

        // If values are empty, continue to next menu_link item.
        if (empty($values['uri'])){
          continue;
        }

        // Check if link is external.
        if ($item->isExternal()) {
          $data['value'][$langcode] = $values['uri'];

          // Set uri type in meta data and continue to next link item.
          $metadata['field'][$event->getFieldName()]['uri_type'] = 'external';
          continue;
        }

        // Explode the uri first by a colon to retreive the link type.
        list($uri_type, $uri_referance) = explode(':', $values['uri'], 2);

        // Set uri type in meta data.
        $metadata['field'][$event->getFieldName()]['uri_type'] = $uri_type;

        // URI handling switch.
        switch ($uri_type) {
          // Entity link.
          case 'entity';
            // Explode entity to get the type and id.
            list($entity_type, $entity_id) = explode('/', $uri_referance, 2);

            // Load the entity to be added as a dependancy.
            $uri_entity = \Drupal::entityTypeManager()->getStorage($entity_type)->load($entity_id);

            // place the entity's uuid into the value position.
            $data['value'][$langcode] = $uri_entity->uuid();
            break;

          // Internal link.
          case 'internal':
            // TODO: CREATE SERIALIZATION OF INTERNAL LINK.
            break;
        }
      }
    }
    // Set data before continuing.
    $event->setFieldData($data);
    // Set the meta data.
    $cdf->setMetadata($metadata);
    // Stop event propagation.
    $event->stopPropagation();
  }

  /**
   * Extract all fields in all translations that match our criteria.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *
   * @return \Drupal\Core\Field\FieldItemListInterface[]
   */
  protected function getFieldsFromEntity(ContentEntityInterface $entity) {
    $fields = [];
    $languages = $entity->getTranslationLanguages();
    /**
     * @var string $field_name
     * @var \Drupal\Core\Field\FieldItemListInterface $field
     */
    foreach ($entity as $field_name => $field) {
      // Check if field definition type is a link.
      if ($field->getFieldDefinition()->getType() === 'link') {
        // If the field is translatable get all translations of it.
        if ($field->getFieldDefinition()->isTranslatable()) {
          foreach ($languages as $language) {
            $translated = $entity->getTranslation($language->getId());
            $fields[] = $translated->get($field_name);
          }
        }
        else {
          $fields[] = $field;
        }
      }
    }
    return $fields;
  }

}
