<?php

namespace Drupal\acquia_contenthub\Form;

use Acquia\ContentHubClient\ContentHubClient;
use Drupal\acquia_contenthub\Client\ClientFactory;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Component\Uuid\Uuid;
use Drupal\Core\Config\ConfigFactoryInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Defines the form to configure the Content Hub connection settings.
 *
 * @todo This is part of the settings work and needs a proper plan.
 */
class ContentHubSettingsForm extends ConfigFormBase {

  const ACH_PATH = '/acquia-contenthub/webhook';

  /**
   * The entity manager.
   *
   * @var |Drupal\acquia_contenthub\Client\ClientFactory
   */
  protected $clientFactory;

  /**
   * The Logger Factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Contenthub Settings
   *
   * @var \Acquia\ContentHubClient\Settings
   */
  protected $settings;

  /**
   * Contenthub Client
   *
   * @var \Acquia\ContentHubClient\ContentHubClient
   */
  protected $client;

  /**
   * Config Factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */

  protected $configFactory;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $dispatcher;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'acquia_contenthub.admin_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['acquia_contenthub.admin_settings'];
  }

  /**
   * ContentHubSettingsForm constructor.
   *
   * @param \Drupal\acquia_contenthub\Client\ClientManagerInterface $client_manager
   *   The client manager.
   * @param \Drupal\acquia_contenthub\ContentHubSubscription $contenthub_subscription
   *   The content hub subscription.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(ConfigFactoryInterface $config_factory, ClientFactory $client_factory, LoggerChannelFactoryInterface $logger_factory, EventDispatcherInterface $dispatcher) {
    $this->configFactory = $config_factory;
    /* The client manager might be null but is still good for later */
    $this->clientFactory = $client_factory;
    $this->loggerFactory = $logger_factory;
    $this->dispatcher = $dispatcher;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('acquia_contenthub.client.factory'),
      $container->get('logger.factory'),
      $container->get('event_dispatcher')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $provider = $this->clientFactory->getProvider();
    $this->settings = $this->clientFactory->getSettings();
    $this->client = $this->clientFactory->getClient();
    $readonly = $provider != 'core_config';
    $connected = !empty($this->client);
    $client_name = '';

    if ($connected) {
      // We need to check remote settings to make sure the name is in sync.
      $client_name = $this->settings->getName();
      $all_clients = $this->client->getRemoteSettings()['clients'];
      foreach ($all_clients AS $client_names) {
        // If its not in-sync secretly change the values and resave the config.
        if ($client_names['uuid'] == $this->settings->getUuid() && $client_names['name'] != $this->settings->getName()) {
          $client_name = $client_names['name'];
          if ($readonly) {
            drupal_set_message($this->t("Warning: client name is out of sync it should be %name. Please manually update your settings file.", ['%name' => $client_name]));
          }
          else {
            $config = $this->config('acquia_contenthub.admin_settings');
            $config->set('client_name', $client_name);
            $config->save();
          }
          continue;
        }
      }
    }

    if ($readonly && $connected) {
      drupal_set_message($this->t("Settings are currently read-only, provided by %provider. Values shown for informational purposes only.", ['%provider' => $provider]));
    }
    else {
      if ($readonly && !$connected) {
        drupal_set_message($this->t("Warning, you are not connected to contenthub, but your settings provider (%provider) does not allow settings changes. Values shown for informational purposes only.", ['%provider' => $provider]), 'warning');
        $connected = TRUE;
      }
    }

    $form['settings'] = [
      '#type' => 'details',
      '#title' => $this->t('Connection Settings'),
      '#collapsible' => TRUE,
      '#description' => $this->t('Settings for connection to Acquia Content Hub'),
      '#disabled' => $connected,
      '#open' => !$connected,
    ];

    $form['settings']['hostname'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Acquia Content Hub Hostname'),
      '#description' => $this->t('The hostname of the Acquia Content Hub API without trailing slash at end of URL, e.g. http://localhost:5000'),
      '#default_value' => $this->settings->getUrl(),
      '#required' => TRUE,
    ];

    $form['settings']['api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API Key'),
      '#default_value' => $this->settings->getApiKey(),
      '#required' => TRUE,
    ];

    $form['settings']['secret_key'] = [
      '#type' => 'password',
      '#title' => $this->t('Secret Key'),
      '#default_value' => $this->settings->getSecretKey(),
      '#required' => TRUE,
    ];

    $form['settings']['client_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Client Name'),
      '#default_value' => $client_name,
      '#required' => TRUE,
      '#description' => $this->t('A unique client name by which Acquia Content Hub will identify this site. The name of this client site cannot be changed once set.'),
    ];

    $form['settings']['webhook'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Publicly Accessible URL'),
      '#required' => TRUE,
      '#default_value' => !$this->settings->getWebhook() ? $GLOBALS['base_url'] : $this->settings->getWebhook(),
      '#description' => 'This should be the domain (and drupal subpath if applicable). Do not include trailing slash.',
    ];

    $form['settings']['origin'] = [
      '#type' => 'item',
      '#title' => $this->t("Site's Origin UUID"),
      '#markup' => $this->settings->getUuid(),
    ];

    if (!$readonly) {
      $form = parent::buildForm($form, $form_state);

      if ($connected) {
        drupal_set_message($this->t("Site successfully connected to contenthub. To change connection settings, unregister the site first."));

        $form['webhook_details'] = [
          '#type' => 'fieldset',
          '#title' => 'Webhook',
          '#collapsible' => FALSE,
          '#open' => TRUE,
        ];
        $form['webhook_details']['webhook'] = $form['settings']['webhook'];
        unset($form['settings']['webhook']);

        $form['actions']['updatewebhook'] = [
          '#type' => 'submit',
          '#value' => $this->t('Update Public URL'),
          '#button_type' => 'primary',
          '#weight' => 100,
          '#submit' => ['::updateWebhook'],
        ];

        $form['actions']['unregister'] = [
          '#type' => 'submit',
          '#value' => $this->t('Unregister Site'),
          '#button_type' => 'secondary',
          '#weight' => 999,
          '#limit_validation_errors' => [],
          '#submit' => ['::unregister'],

          // No validation at all is required in the equivocate case, so
          // we include this here to make it skip the form-level validator.
          '#validate' => [],
        ];
        unset($form['actions']['submit']);

      }
      else {
        $form['actions']['submit']['#value'] = $this->t('Register Site');
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $hostname = NULL;
    if ($form_state->getTriggeringElement()['#id'] == 'edit-unregister') {
      return;
    }
    // Our submit button is updating a public URL
    // Validate that the webhook exists and is accessible
    if (UrlHelper::isValid($form_state->getValue('webhook'), TRUE)) {
      $webhook_url = $form_state->getValue('webhook');
    }
    else {
      return $form_state->setErrorByName('webhook', $this->t('Please type a publically accessible url.'));
    }
    if ($form_state->getTriggeringElement()['#id'] == 'edit-updatewebhook') {
      // Because we must delete and regenerate the webhook UUID, don't allow
      // the site to take over an existing webhook url.
      if (!empty($this->client->getWebHook($webhook_url))) {
        $message = $this->t('The webhook url "%name" is already being used. Please insert another one, or unregister the existing one first.', [
        '%name' => $webhook_url,
        ]);
        return $form_state->setErrorByName('webhook', $message);
      }
      return;
    }

    if (UrlHelper::isValid($form_state->getValue('hostname'), TRUE)) {
      // Remove trailing slash at end of URL.
      $hostname = rtrim($form_state->getValue('hostname'), '/');
    }
    else {
      return $form_state->setErrorByName('hostname', $this->t('This is not a valid URL. Please insert it again.'));
    }

    $api = NULL;
    // Important. This should never validate if it is an UUID. Lift 3 does not
    // use UUIDs for the api_key but they are valid for Content Hub.
    if ($form_state->getValue('api_key')) {
      $api = $form_state->getValue('api_key');
    }
    else {
      return $form_state->setErrorByName('api_key', $this->t('Please insert an API Key.'));
    }

    $secret = NULL;
    if ($form_state->hasValue('secret_key')) {
      $secret = $form_state->getValue('secret_key');
    }
    else {
      return $form_state->setErrorByName('secret_key', $this->t('Please insert a Secret Key.'));
    }

    if ($form_state->hasValue('client_name')) {
      $client_name = $form_state->getValue('client_name');
    }
    else {
      return $form_state->setErrorByName('client_name', $this->t('Please insert a Client Name.'));
    }

    if (Uuid::isValid($form_state->getValue('origin'))) {
      $origin = $form_state->getValue('origin');
    }
    else {
      $origin = '';
    }

    // First Register the site to Content Hub.
    $client_name = $form_state->getValue('client_name');
    $logger_channel = $this->loggerFactory->get('acquia_contenthub');
    try {
      if ($client = ContentHubClient::register($logger_channel, $this->dispatcher, $client_name, $form_state->getValue('hostname'), $form_state->getValue('api_key'), $form_state->getValue('secret_key'))) {
        // If the client passes back, then we're good.
        $this->client = $client;
      }
    } catch (\GuzzleHttp\Exception\RequestException $ch_error) {
      //Were we able to connect to the serivce?
      if ($ch_error->getResponse()) {
        $service_error = json_decode($ch_error->getResponse()->getBody(), TRUE);
        $message = $service_error['error']['message'];
      }
      // We failed during the response, grab the parent request exception.
      else {
        $service_error['error']['code'] = 1;
        $message = $ch_error->getMessage();
      }

      if (isset($service_error['error']['code'])) {
        switch ($service_error['error']['code']) {
          case 4006:
            $form_state->setErrorByName('client_name', $message);
            break;
          case 401:
            $form_state->setErrorByName('secret_key', $this->t("Access Denied, Reason: %message", ["%message" => $message]));
            break;
          case 1:
            $form_state->setErrorByName('hostname', $message);
          default:
            $message = $this->t('There is a problem connecting to Acquia Content Hub. Please ensure that your hostname and credentials are correct.');
            $form_state->setErrorByName('settings', $message);
        }
      }
    }
  }

  public function updateWebhook(array &$form, FormStateInterface $form_state) {
    $config = $this->config('acquia_contenthub.admin_settings');
    $webhook_url = rtrim($form_state->getValue('webhook'), "/") . $this::ACH_PATH;

    // Delete the existing webhook first. To be removed once updatewebhook functions
    $delete_response = $this->client->deleteWebhook($this->settings->getWebhook('uuid'));

    // Add the new webhook.
    $response = $this->client->addWebhook($webhook_url);

    /**
     * The following code is for update webhook, which doesn't work yet. Leaving
     * Here so we can uncomment it once the plexus fix is in.
     * $response = $this->client->updateWebhook($this->settings->getWebhook('uuid'), $webhook_url);
     */
    // Yay we recieved a response. Is it valid?
    if (empty($response) || (isset($response['success']) && $response['success'] == FALSE)) {
      // Webhook already exists, just update the UUID to match plexus.
      if (isset($response['error']['code']) && $response['error']['code'] == 4010) {
        $response = $this->client->getWebHook($webhook_url);
      }
      else if (isset($response['error']['message'])) {
        drupal_set_message($this->t("Unable to create URL %url, Error %error: " . $response['error']['message'], [
          '%url' => $webhook_url,
          '%error' => $response['error']['code']
        ]), 'error');
        return;
      }
      else {
        drupal_set_message($this->t("Unable to create URL %url, Unable to connect to Contenthub."), 'error');
        return;
      }
    }

    $webhook = [
      'uuid' => $response['uuid'],
      'url' => $response['url'],
      'settings_url' => rtrim($form_state->getValue('webhook'), "/")
    ];
    $config->set('webhook', $webhook);
    $config->save();
    drupal_set_message($this->t("Successfully Updated Public URL."));

  }

  public function unregister(array &$form, FormStateInterface $form_state) {
    /*
    $webhook_register = (bool) $this->config->get('webhook_uuid');
    $webhook_url = $this->config->get('webhook_url');
    // Un-register the webhook.
    if ($webhook_register) {
      $this->unregisterWebhook($webhook_url);
    }*/
    $client = $this->clientFactory->getClient();
    try {
      $client->deleteClient();
    }
    catch (\Exception $e) {
      $deleted = true;
    }

    $config = $this->config('acquia_contenthub.admin_settings');
    $client_name = $config->get('client_name');
    $config->delete();
    drupal_set_message($this->t("Successfully disconnected site %site from contenthub", ['%site' => $client_name]));
    Cache::invalidateTags(['acquia_contenthub_settings']);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);

    // We assume here all inserted values have passed validation.
    $config = $this->config('acquia_contenthub.admin_settings');
    $this->settings = $this->client->getSettings();
    if ($this->settings->getUrl()) {
      // Remove trailing slash at end of URL.
      $hostname = rtrim($this->settings->getUrl(), '/');
      $config->set('hostname', $hostname);
    }

    if ($this->settings->getApiKey()) {
      $config->set('api_key', $this->settings->getApiKey());
    }

    if ($this->settings->getSecretKey()) {
      $config->set('secret_key', $this->settings->getSecretKey());
    }

    if ($this->settings->getUuid()) {
      $config->set('origin', $this->settings->getUuid());
    }

    if ($this->settings->getName()) {
      $config->set('client_name', $this->settings->getName());
    }

    $remote = $this->client->getRemoteSettings();
    if ($remote['shared_secret']) {
      $config->set('shared_secret', $remote['shared_secret']);
    }

    // Call Save now, so when the webhook call from Plexus occurs, we can
    // respond with a valid signature.
    $config->save();
    Cache::invalidateTags(['acquia_contenthub_settings']);

    // If a webhook was sent during registration, lets get that added as
    // well, return the webhook UUID if successful.
    $webhook = $form_state->getValue('webhook');
    if (!empty($webhook)) {
      $webhook_response = $this->client->addWebhook(rtrim($webhook, "/") . $this::ACH_PATH);
      if (isset($webhook_response['success']) && $webhook_response['success'] == FALSE) {
        // If the webhook already exists on the subscription, Grab its data.
        if (isset($webhook_response['error']['code']) && $webhook_response['error']['code'] == 4010) {
          $webhook_response = $this->client->getWebHook(rtrim($webhook, "/") . $this::ACH_PATH);
        }
        else {
          drupal_set_message((sprintf("Unable to create URL %s, Error %d: %s", $webhook, $webhook_response['error']['code'], $webhook_response['error']['message'])));
          drupal_set_message("Registration to Contenthub was successful, but Contenthub could not reach your site to verify connectivity. Please update your publically accessible URL and try again.");
          return;
        }
      }
      $webhook = ['uuid' => $webhook_response['uuid'], 'url' => $webhook_response['url'], 'settings_url' => rtrim($webhook, "/")];
      $config->set('webhook', $webhook);
      $config->save();
    }
  }
}
